import hashlib
import typing as tp
import mimetypes
import os
import datetime
import uuid

from django.core.files.uploadedfile import UploadedFile
from django.db import models
from django.utils.safestring import mark_safe
from django_common.auth_backends import User

from satella.coding import silence_excs
from satella.instrumentation.metrics import getMetric

from netguru.settings import FILE_SAVE_LOCATION, SECRET_KEY

files_deleted = getMetric('deleted.files', 'counter')
links_deleted = getMetric('deleted.links', 'counter')
files_created = getMetric('created.files', 'counter')
links_created = getMetric('created.links', 'counter')

SHARE_URL = 0
SHARE_FILE = 1

SHARE_TYPES = [
    (SHARE_URL, 'URL'),
    (SHARE_FILE, 'File')
]


class ShareManager(models.Manager):
    def get_for_day(self, date: datetime.date):
        return super().get_queryset().filter(created_on__gte=date,
                                             created_on__lt=date + datetime.timedelta(days=1))


class Share(models.Model):
    """
    A share is a single shared resource.

    It is created on being added and removed upon the reaper task collecting it.
    """
    creator = models.ForeignKey('auth.User', verbose_name='Creator', on_delete=models.CASCADE)
    share_type = models.IntegerField(choices=SHARE_TYPES, verbose_name='Share type')
    created_on = models.DateTimeField(default=datetime.datetime.now,
                                      verbose_name='Created on',
                                      db_index=True)
    pwd_hash = models.CharField(max_length=64, verbose_name='Hashed password salt')
    resource = models.TextField(verbose_name='Resource',
                                help_text='URL if this is an URL. Otherwise it will be '
                                          'name of the file as it is on filesystem, then a dot,'
                                          'and then the file name the user submitted it as')
    times_used = models.IntegerField(verbose_name='Visit count', default=0)

    objects = ShareManager()

    def get_absolute_url(self) -> str:
        return f'/shares/{self.id}'

    def __str__(self):
        if self.share_type == SHARE_URL:
            return f'Link to {self.resource} no {self.id}'
        else:
            return f'File {self.file_name} no {self.id}'

    @property
    def creation_date(self) -> datetime.date:
        return datetime.date(year=self.created_on.year,
                             month=self.created_on.month,
                             day=self.created_on.day)

    @property
    def expired(self) -> bool:
        """Has the link expired?"""
        return self.created_on < datetime.datetime.now() - datetime.timedelta(days=1)

    @property
    def mimetype(self) -> str:
        """Guessed mimetype or application/octet-stream"""
        return mimetypes.guess_type(self.file_name)[0] or 'application/octet-stream'

    def get_kwargs_for_s_http_response(self) -> tp.Dict:
        """
        Return self as set of kwargs for streaming http response
        """
        return {
            'streaming_content': self.file_iterator,
            'content_type': self.mimetype,
            'headers': {
                'Content-Disposition': f'attachment; filename="{self.file_name}"'
            }
        }

    @property
    def file_iterator(self) -> tp.Iterator[bytes]:
        """
        Return a file iterator.
        Must be a SHARE_FILE
        """
        with open(self.path, 'rb') as f_in:
            while p := f_in.read(1024):
                yield p

    @property
    def file_name(self) -> str:
        """
        Return real file name (as submitted by the user).
        Must be a SHARE_FILE
        """
        assert self.share_type == SHARE_FILE, 'Not a file!'
        return self.resource.split('.', 1)[1]

    @property
    def uuid_name(self) -> str:
        """Return file name. Must be a SHARE_FILE"""
        assert self.share_type == SHARE_FILE, 'Not a file!'
        return self.resource.split('.', 1)[0]

    @property
    def path(self) -> str:
        """Return path to file"""
        return os.path.join(FILE_SAVE_LOCATION, self.uuid_name)

    def delete(self, *args, **kwargs):
        if self.share_type == SHARE_FILE:
            path = os.path.join(FILE_SAVE_LOCATION, self.uuid_name)
            # Since we might have already deleted it, but our transaction was rolled
            # back because of reasons.
            # This is both EAFP (more Pythonic) and shorter 2 lines than a try ... catch
            # besides it's a great showcase of my Satella
            with silence_excs(FileNotFoundError):
                os.unlink(path)
            files_deleted.runtime(+1)
        else:
            links_deleted.runtime(+1)
        super().delete(*args, **kwargs)


def hash_password(password: str) -> str:
    return hashlib.sha256(password.encode('utf-8') + SECRET_KEY.encode('utf-8')).hexdigest()


def generate_correct_uuid() -> str:
    f = uuid.uuid4().hex
    # I know that UUID collisions practically don't happen, but since it's so cheap for us
    # to go that extra mile
    # and since we're at risk at overwriting somebody's files, I guess i should check it
    # I realize that it's susceptible to race conditions, but then again what is the chance?
    # The chance that UUID will repeat during the 24 hours a link is active is much higher
    # than the chance that it will repeat during the same 2 second window
    while os.path.exists(os.path.join(FILE_SAVE_LOCATION, f)):
        f = uuid.uuid4().hex
    return f


def add_file(file: UploadedFile, request, escape_url: bool = True) -> dict:
    """
    Adds a new file

    :param file: file to add
    :param request: request to use
    :return: a dictionary with following keys (password, url, type)
    """
    data_added = {'password': User.objects.make_random_password(), 'type': 'file'}
    pwd_hash = hash_password(data_added['password'])
    file_name = generate_correct_uuid()
    resource_name = f'{file_name}.{file.name}'
    with open(os.path.join(FILE_SAVE_LOCATION, file_name), 'wb') as f_out:
        for chunk in file.chunks():
            f_out.write(chunk)
    share = Share(creator=request.user,
                  resource=resource_name,
                  pwd_hash=pwd_hash,
                  share_type=SHARE_FILE)
    files_created.runtime(+1)
    share.save()
    url = f'https://{request.get_host()}/shares/{share.id}'
    if escape_url:
        url = mark_safe(url)
    data_added['url'] = url
    return data_added


def add_url(url: str, request, escape_url: bool = True) -> dict:
    """
    Adds a new URL

    :param file: file to add
    :param request: request to use
    :return: a dictionary with following keys (password, url, type)
    """
    data_added = {'password': User.objects.make_random_password(), 'type': 'file'}
    pwd_hash = hash_password(data_added['password'])
    share = Share(creator=request.user,
                  resource=url,
                  pwd_hash=pwd_hash,
                  share_type=SHARE_URL)
    links_created.runtime(+1)
    share.save()
    url = f'https://{request.get_host()}/shares/{share.id}'
    if escape_url:
        url = mark_safe(url)
    data_added['url'] = url
    return data_added
