import mimetypes
import typing as tp
import os
import random
import logging
import string
import uuid
from django import forms
from django.contrib.auth.models import User
from django.core.validators import URLValidator
from django.forms import fields
from django.core.exceptions import ValidationError
from django.http import Http404, StreamingHttpResponse as SHTTPResponse
from django.shortcuts import render, get_object_or_404, redirect
from django.contrib.auth.decorators import login_required
from django.utils.safestring import mark_safe
from satella.instrumentation.metrics import getMetric

from shares.models import Share, SHARE_URL, SHARE_FILE, hash_password, add_file, add_url
import hashlib
from netguru.settings import SECRET_KEY, FILE_SAVE_LOCATION

files = getMetric('visited.form.files', 'counter')
links = getMetric('visited.form.links', 'counter')
logger = logging.getLogger(__name__)


#  ==============
#  Add share part
#  ==============


class AddShareURLForm(forms.Form):
    url = fields.CharField(label='URL', required=False,
                           validators=[URLValidator()])
    file = fields.FileField(label='File', required=False)

    def clean(self):
        cleaned_data = super().clean()
        if not cleaned_data.get('url') and not cleaned_data.get('file'):
            raise ValidationError('You have to provide either an URL or a file!')
        return cleaned_data


@login_required
def add_share(request):
    data_added = None

    if request.method == 'POST':
        form = AddShareURLForm(request.POST, request.FILES)
        if form.is_valid():

            pwd_hash = hash_password(data_added['password'])
            data = form.cleaned_data
            if data.get('url'):
                # Create a URL resource
                data_added = add_url(data['url'], request)
                logger.info('Created a link via form')
            else:
                data_added = add_file(request.FILES['file'], request)
                logger.info('Created a file via form')
    else:
        form = AddShareURLForm()

    return render(request, 'share/add.html', {'form': form, 'added': data_added})


#  ===============
#  View share part
#  ===============

class PasswordForm(forms.Form):
    password = fields.CharField(label='Password', widget=forms.PasswordInput)


class InvalidPassword(Exception):
    ...


def view_share(request, share_id: str):
    def render_me(arg, **kwargs):
        return render(request, 'share/view.html', arg, **kwargs)

    try:
        share = get_object_or_404(Share, id=share_id)

        if share.expired:
            raise Http404()

        form = PasswordForm()
        if request.method == 'POST':
            form = PasswordForm(request.POST)
            if form.is_valid():
                pwd = form.cleaned_data['password']
                if hash_password(pwd) != share.pwd_hash:
                    raise InvalidPassword()
                try:
                    if share.share_type == SHARE_URL:
                        links.runtime(+1)
                        return redirect(share.resource)
                    else:
                        if not os.path.exists(share.path):
                            raise Http404()

                        # Django will have the UTF-8 mumbo jumbo for us
                        files.runtime(+1)
                        return SHTTPResponse(**share.get_kwargs_for_s_http_response())
                finally:
                    share.times_used += 1
                    share.save()
    except Http404:
        return render_me({'not_found': True, 'share_id': share_id}, status=404)
    except InvalidPassword:
        return render_me({'form': form, 'invalid_password': True, 'share_id': share_id}, status=401)

    return render_me({'form': form, 'share_id': share_id})
